import numpy as np
import torch 
from torchtuples.tupletree import TupleTree, tuplefy


def assert_tupletree_equal(a, b, check_dtypes=True):
    assert type(a) == type(b) == TupleTree, 'Not TupleTree'
    assert a.numerate() == b.numerate(), 'Not same structure'
    assert a.types() == b.types(), 'Not same types'
    if check_dtypes:
        ad, bd = (tuplefy(a, b)
                  .apply(lambda x: x.dtype if hasattr(x, 'dtype') else 'not_tensor'))
        assert ad == bd, 'Not same dtype'

    for aa, bb in zip(a.flatten(), b.flatten()):
        if hasattr(aa, 'dtype'):
            assert (aa == bb).all(), 'Not equal values'
        else:
            assert aa == bb, 'Not equal values'


"""Model for fitting torch models.
"""
import os
from collections import OrderedDict, defaultdict
import warnings
import contextlib
from typing import Dict
import numpy as np
import torch
import torchtuples.callbacks as cb
from torchtuples.optim import AdamW, OptimWrap
from torchtuples.tupletree import tuplefy, TupleTree, make_dataloader
from torchtuples.utils import make_name_hash, array_or_tensor, is_data, is_dl


class Model(object):
    """Train torch models using dataloaders, tensors or np.arrays.

    Arguments:
        net {torch.nn.Module} -- A torch module.

    Keyword Arguments:
        loss {function} -- Set function that is used for training
            (e.g. binary_cross_entropy for torch) (default: {None})
        optimizer {Optimizer} -- A torch optimizer or similar. Preferrably use torchtuples.optim instead of
            torch.optim, as this allows for reinitialization, etc. If 'None' set to torchtuples.optim.AdamW.
            (default: {None})
        device {str, int, torch.device} -- Device to compute on. (default: {None})
            Preferrably pass a torch.device object.
            If 'None': use default gpu if avaiable, else use cpu.
            If 'int': used that gpu: torch.device('cuda:<device>').
            If 'string': string is passed to torch.device('string').

    Example simple model:
    ---------------------
    from torchtuples import Model
    import torch
    from torch import nn
    from torch.nn import functional as F

    n_rows, n_covs = 1000, 4
    x = torch.randn((n_rows, n_covs))
    y = 2 * x.sum(1, keepdim=True) + 4  # y = 2 * x + 4

    net = nn.Sequential(nn.Linear(n_covs, 10), nn.ReLU(), nn.Linear(10, 1))
    loss = F.mse_loss
    model = Model(net, loss)
    log = model.fit(x, y, batch_size=32, epochs=30)
    log.plot()
    """

    def __init__(self, net, loss=None, optimizer=None, device=None):
        self.net = net
        if type(self.net) is str:
            self.load_net(self.net)
        self.loss = loss
        self.optimizer = optimizer if optimizer is not None else AdamW
        self.set_device(device)
        if not hasattr(self, "make_dataloader_predict"):
            self.make_dataloader_predict = self.make_dataloader
        self._init_train_log()
        self.metrics = self._setup_metrics()

    def _init_train_log(self):
        self.log = cb.TrainingLogger()
        self.train_metrics = cb._MonitorFitMetricsTrainData()
        self.val_metrics = cb.MonitorFitMetrics()
        self.log.monitors = OrderedDict(train_=self.train_metrics, val_=self.val_metrics)
        self.callbacks = None

    @property
    def device(self):
        return self._device

    def set_device(self, device):
        """Set the device used by the model.
        This is called in the __init__ function, but can be used to later change the device.

        Arguments:
            device {str, int, torch.device} -- Device to compute on. (default: {None})
                Preferrably pass a torch.device object.
                If 'None': use default gpu if avaiable, else use cpu.
                If 'int': used that gpu: torch.device('cuda:<device>').
                If 'string': string is passed to torch.device('string').
        """
        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        elif type(device) is str:
            device = torch.device(device)
        elif type(device) is int:
            device = torch.device(f"cuda:{device}")
        if type(device) is not torch.device:
            raise ValueError(
                "Argument `device` needs to be `None`, `string`, `int`, or `torch.device`,"
                + f" got {type(device)}"
            )
        self._device = device
        self.net.to(self.device)

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer):
        self._optimizer = optimizer
        if callable(self._optimizer):
            self._optimizer = self._optimizer(params=self.net.parameters())
        if not isinstance(self._optimizer, OptimWrap):
            self._optimizer = OptimWrap(self._optimizer)

    @staticmethod
    def make_dataloader(data, batch_size, shuffle, num_workers=0, **kwargs):
        """Function for creating a dataloader from tensors or arrays.
        It is natural to rewrite this method in inherited classes.

        self.make_dataloader_predict will be set to this method if not implemented
        separatelly.

        This simply calls tupletree.make_dataloader, but is included to make
        inheritance simpler.

        Arguments:
            data {tuple, np.array, tensor} -- Data in dataloader e.g. (x, y)
            batch_size {int} -- Batch size used in dataloader
            shuffle {bool} -- If order should be suffled

        Keyword Arguments:
            num_workers {int} -- Number of workers in dataloader (default: {0})
            to_tensor {bool} -- Ensure that we use tensors (default: {True})
            **kwargs -- Passed to make_dataloader.

        Returns:
            DataLoaderBatch -- A dataloader object like the torch DataLoader
        """
        dataloader = make_dataloader(data, batch_size, shuffle, num_workers, **kwargs)
        return dataloader

    def _setup_train_info(self, dataloader):
        self.fit_info = {"batches_per_epoch": len(dataloader)}

        def _tuple_info(tuple_):
            tuple_ = tuplefy(tuple_)
            return {"levels": tuple_.to_levels(), "shapes": tuple_.shapes().apply(lambda x: x[1:])}

        data = _get_element_in_dataloader(dataloader)
        if data is not None:
            if len(data) == 2:
                try:
                    input, target = data
                    self.fit_info["input"] = _tuple_info(input)
                except:
                    pass

    def _to_device(self, data) -> TupleTree:
        """Move `data` to self.device.
        If `data` is a tensor, it will be returned as a `TupleTree`.
        """
        if data is None:
            return tuplefy(data)
        return tuplefy(data).to_device(self.device)

    def compute_metrics(self, data, metrics=None) -> Dict[str, torch.Tensor]:
        """Function for computing the loss and other metrics.

        Arguments:
            data {tensor or tuple} -- A batch of data. Typically the tuple `(input, target)`.

        Keyword Arguments:
            metrics {dict} -- A dictionary with metrics. If `None` use `self.metrics`. (default: {None})
        """
        if metrics is None:
            metrics = self.metrics
        if (self.loss is None) and (self.loss in metrics.values()):
            raise RuntimeError(f"Need to set `self.loss`.")

        input, target = data
        input = self._to_device(input)
        target = self._to_device(target)
        out = self.net(*input)
        out = tuplefy(out)
        return {name: metric(*out, *target) for name, metric in metrics.items()}

    def _setup_metrics(self, metrics=None):
        all_metrics = {"loss": self.loss}
        if metrics is not None:
            if not hasattr(metrics, "items"):
                if not hasattr(metrics, "__iter__"):
                    metrics = [metrics]
                metrics = {met.__name__: met for met in metrics}
            if "loss" in metrics:
                raise ValueError("The 'loss' keyword is reserved for the loss function.")
            all_metrics.update(metrics)
        return all_metrics

    def fit_dataloader(
        self, dataloader, epochs=1, callbacks=None, verbose=True, metrics=None, val_dataloader=None
    ):
        """Fit a dataloader object.
        See 'fit' for tensors and np.arrays.

        Arguments:
            dataloader {dataloader} -- A dataloader that gives (input, target).

        Keyword Arguments:
            epochs {int} -- Number of epochs (default: {1})
            callbacks {list} -- list of callbacks (default: {None})
            verbose {bool} -- Print progress (default: {True})

        Returns:
            TrainingLogger -- Training log
        """
        self._setup_train_info(dataloader)
        self.metrics = self._setup_metrics(metrics)
        self.log.verbose = verbose
        self.val_metrics.dataloader = val_dataloader
        if callbacks is None:
            callbacks = []
        self.callbacks = cb.TrainingCallbackHandler(
            self.optimizer, self.train_metrics, self.log, self.val_metrics, callbacks
        )
        self.callbacks.give_model(self)

        stop = self.callbacks.on_fit_start()
        for _ in range(epochs):
            if stop:
                break
            stop = self.callbacks.on_epoch_start()
            if stop:
                break
            for data in dataloader:
                stop = self.callbacks.on_batch_start()
                if stop:
                    break
                self.optimizer.zero_grad()
                self.batch_metrics = self.compute_metrics(data, self.metrics)
                self.batch_loss = self.batch_metrics["loss"]
                self.batch_loss.backward()
                stop = self.callbacks.before_step()
                if stop:
                    break
                self.optimizer.step()
                stop = self.callbacks.on_batch_end()
                if stop:
                    break
            else:
                stop = self.callbacks.on_epoch_end()
        self.callbacks.on_fit_end()
        return self.log

    def fit(
        self,
        input,
        target=None,
        batch_size=256,
        epochs=1,
        callbacks=None,
        verbose=True,
        num_workers=0,
        shuffle=True,
        metrics=None,
        val_data=None,
        val_batch_size=8224,
        **kwargs,
    ):
        """Fit  model with inputs and targets.

        Arguments:
            input {np.array, tensor or tuple} -- Input (x) passed to net.
            target {np.array, tensor or tuple} -- Target (y) passed to loss function.

        Keyword Arguments:
            batch_size {int} -- Elemets in each batch (default: {256})
            epochs {int} -- Number of epochs (default: {1})
            callbacks {list} -- list of callbacks (default: {None})
            verbose {bool} -- Print progress (default: {True})
            num_workers {int} -- Number of workers used in the dataloader (default: {0})
            shuffle {bool} -- If we should shuffle the order of the dataset (default: {True})
            **kwargs -- Passed to the 'make_dataloader' method. Set e.g. `torch_ds_dl to use
                the TensorDataset and DataLoader provided by torch instead of the torchtuples
                implementations.

        Returns:
            TrainingLogger -- Training log
        """
        if target is not None:
            input = (input, target)
        dataloader = self.make_dataloader(input, batch_size, shuffle, num_workers, **kwargs)
        val_dataloader = val_data
        if (is_dl(val_data) is False) and (val_data is not None):
            val_dataloader = self.make_dataloader(
                val_data, val_batch_size, shuffle=False, num_workers=num_workers, **kwargs
            )
        log = self.fit_dataloader(dataloader, epochs, callbacks, verbose, metrics, val_dataloader)
        return log

    @contextlib.contextmanager
    def _lr_finder(self, lr_min, lr_max, lr_range, n_steps, tolerance, verbose):
        lr_lower, lr_upper = lr_range
        path = make_name_hash("lr_finder_checkpoint")
        self.save_model_weights(path)
        try:
            self.optimizer.drop_scheduler()
            lr_finder = cb.LRFinder(lr_lower, lr_upper, n_steps, tolerance)
            yield lr_finder
        except Exception as e:
            self.load_model_weights(path)
            os.remove(path)
            self.optimizer = self.optimizer.reinitialize()
            self._init_train_log()
            raise e
        self.load_model_weights(path)
        os.remove(path)
        lr = lr_finder.get_best_lr(lr_min, lr_max)
        self.optimizer = self.optimizer.reinitialize(lr=lr)
        self._init_train_log()

    def lr_finder(
        self,
        input,
        target,
        batch_size=256,
        lr_min=1e-4,
        lr_max=1.0,
        lr_range=(1e-7, 10.0),
        n_steps=100,
        tolerance=np.inf,
        callbacks=None,
        verbose=False,
        num_workers=0,
        shuffle=True,
        **kwargs,
    ):
        with self._lr_finder(lr_min, lr_max, lr_range, n_steps, tolerance, verbose) as lr_finder:
            if callbacks is None:
                callbacks = []
            callbacks.append(lr_finder)
            epochs = n_steps
            self.fit(
                input,
                target,
                batch_size,
                epochs,
                callbacks,
                verbose,
                num_workers,
                shuffle,
                **kwargs,
            )
        return lr_finder

    def lr_finder_dataloader(
        self,
        dataloader,
        lr_min=1e-4,
        lr_max=1.0,
        lr_range=(1e-7, 10.0),
        n_steps=100,
        tolerance=np.inf,
        callbacks=None,
        verbose=False,
    ):
        with self._lr_finder(lr_min, lr_max, lr_range, n_steps, tolerance, verbose) as lr_finder:
            if callbacks is None:
                callbacks = []
            callbacks.append(lr_finder)
            epochs = n_steps
            self.fit_dataloader(dataloader, epochs, callbacks, verbose)
        return lr_finder

    def score_in_batches(
        self,
        input,
        target=None,
        score_func=None,
        batch_size=8224,
        eval_=True,
        mean=True,
        num_workers=0,
        shuffle=False,
        make_dataloader=None,
        numpy=True,
        **kwargs,
    ):
        """Used to score a dataset in batches.
        If score_func is None, this use the loss function.
        If make_dataloader is None, we use self.make_dataloader_predict, unless score_func is also
        None, in which we use self.make_dataloader.

        Arguments:
            data {np.array, tensor, tuple, dataloader} -- Data in the form a datloader, or arrarys/tensors.

        Keyword Arguments:
            score_func {func} -- Function used for scoreing. If None, we use self.loss. (default: {None})
            batch_size {int} -- Batch size (default: {8224})
            eval_ {bool} -- Eval mode of the net. (default: {True})
            mean {bool} -- If True, we return the mean. (default: {True})
            num_workers {int} -- Number of workers for the dataloader. (default: {0})
            shuffle {bool} -- If the data should be shuffled (default: {False})
            make_dataloader {func} -- Function for making a dataloder.
                If None, we use make_dataloader_predict as long as score_func is not None. (default: {None})
            **kwargs -- Are passed to make_dataloader function.

        Returns:
            np.array -- Scores
        """
        if make_dataloader is None:
            if score_func is None:
                make_dataloader = self.make_dataloader
            else:
                make_dataloader = self.make_dataloader_predict
        if target is not None:
            input = (input, target)
        dl = make_dataloader(input, batch_size, shuffle, num_workers, **kwargs)
        scores = self.score_in_batches_dataloader(dl, score_func, eval_, mean, numpy)
        return scores

    def score_in_batches_dataloader(
        self, dataloader, score_func=None, eval_=True, mean=True, numpy=True
    ):
        """Score a dataset in batches.

        Parameters:
            dataloader: Dataloader:
            score_func: Function of (self, data) that returns a measure.
                If None, we get training loss.
            eval_: If net should be in eval mode.
            mean: If return mean or list with scores.
        """
        if eval_:
            self.net.eval()
        batch_scores = []
        with torch.no_grad():
            for data in dataloader:
                if score_func is None:
                    score = self.compute_metrics(data, self.metrics)
                else:
                    warnings.warn(
                        f"score_func {score_func} probably doesn't work... Not implemented"
                    )
                    input, target = data
                    score = score_func(self, input, target)
                batch_scores.append(score)
        if eval_:
            self.net.train()
        if type(batch_scores[0]) is dict:
            scores = defaultdict(list)
            for bs in batch_scores:
                for name, score in bs.items():
                    scores[name].append(score)
            scores = {name: torch.tensor(score) for name, score in scores.items()}
            if mean:
                scores = {name: score.mean() for name, score in scores.items()}
            if numpy:
                scores = {name: score.item() for name, score in scores.items()}
            return scores
        if mean:
            batch_scores = [score.item() for score in batch_scores]
            return np.mean(batch_scores)
        return batch_scores

    def _predict_func_dl(
        self, func, dataloader, numpy=False, eval_=True, grads=False, to_cpu=False
    ):
        """Get predictions from `dataloader`.
        `func` can be anything and is not concatenated to `self.net` or `self.net.predict`.
        This is different from `predict` and `predict_net` which both use call `self.net`.
        """
        if hasattr(self, "fit_info") and (self.make_dataloader is self.make_dataloader_predict):
            data = _get_element_in_dataloader(dataloader)
            if data is not None:
                input = tuplefy(data)
                input_train = self.fit_info["input"]
                if input.to_levels() != input_train["levels"]:
                    warnings.warn(
                        """The input from the dataloader is different from
                    the 'input' during trainig. Make sure to remove 'target' from dataloader.
                    Can be done with 'torchtuples.data.dataloader_input_only'."""
                    )
                if input.shapes().apply(lambda x: x[1:]) != input_train["shapes"]:
                    warnings.warn(
                        """The input from the dataloader is different from
                    the 'input' during trainig. The shapes are different."""
                    )

        if eval_:
            self.net.eval()
        with torch.set_grad_enabled(grads):
            preds = []
            for input in dataloader:
                input = tuplefy(input).to_device(self.device)
                preds_batch = tuplefy(func(*input))
                if numpy or to_cpu:
                    preds_batch = preds_batch.to_device("cpu")
                preds.append(preds_batch)
        if eval_:
            self.net.train()
        preds = tuplefy(preds).cat()
        if numpy:
            preds = preds.to_numpy()
        if len(preds) == 1:
            preds = preds[0]
        return preds

    def _predict_func(
        self,
        func,
        input,
        batch_size=8224,
        numpy=None,
        eval_=True,
        grads=False,
        to_cpu=False,
        num_workers=0,
        is_dataloader=None,
        **kwargs,
    ):
        """Get predictions from `input` which can be data or a DataLoader.
        `func` can be anything and is not concatenated to `self.net` or `self.net.predict`.
        This is different from `predict` and `predict_net` which both use call `self.net`.
        """
        if is_data(input) or (is_dataloader is False):
            dl = self.make_dataloader_predict(
                input, batch_size, shuffle=False, num_workers=num_workers, **kwargs
            )
        elif is_dl(input) or (is_dataloader is True):
            dl = input
        else:
            raise ValueError(
                "Did not recognize data type. You can set `is_dataloader to `True`"
                + " or `False` to force usage."
            )

        to_cpu = numpy or to_cpu
        preds = self._predict_func_dl(func, dl, numpy, eval_, grads, to_cpu)
        return array_or_tensor(preds, numpy, input)

    def predict_net(
        self,
        input,
        batch_size=8224,
        numpy=None,
        eval_=True,
        grads=False,
        to_cpu=False,
        num_workers=0,
        is_dataloader=None,
        func=None,
        **kwargs,
    ):
        """Get predictions from 'input' using the `self.net(x)` method.
        Use `predict` instead if you want to use `self.net.predict(x)`.

        Arguments:
            input {dataloader, tuple, np.ndarra, or torch.tensor} -- Input to net.

        Keyword Arguments:
            batch_size {int} -- Batch size (default: {8224})
            numpy {bool} -- 'False' gives tensor, 'True' gives numpy, and None give same as input
                (default: {None})
            eval_ {bool} -- If 'True', use 'eval' modede on net. (default: {True})
            grads {bool} -- If gradients should be computed (default: {False})
            to_cpu {bool} -- For larger data sets we need to move the results to cpu
                (default: {False})
            num_workers {int} -- Number of workes in created dataloader (default: {0})
            func {func} -- A toch function, such as `torch.sigmoid` which is called after the predict.
                (default: {None})
            **kwargs -- Passed to make_dataloader.

        Returns:
            [TupleTree, np.ndarray or tensor] -- Predictions
        """
        pred_func = wrapfunc(func, self.net)
        preds = self._predict_func(
            pred_func,
            input,
            batch_size,
            numpy,
            eval_,
            grads,
            to_cpu,
            num_workers,
            is_dataloader,
            **kwargs,
        )
        return array_or_tensor(preds, numpy, input)

    def predict(
        self,
        input,
        batch_size=8224,
        numpy=None,
        eval_=True,
        grads=False,
        to_cpu=False,
        num_workers=0,
        is_dataloader=None,
        func=None,
        **kwargs,
    ):
        """Get predictions from 'input' using the `self.net.predict(x)` method.
        Use `predict_net` instead if you want to use `self.net(x)`.

        Arguments:
            input {dataloader, tuple, np.ndarra, or torch.tensor} -- Input to net.

        Keyword Arguments:
            batch_size {int} -- Batch size (default: {8224})
            numpy {bool} -- 'False' gives tensor, 'True' gives numpy, and None give same as input
                (default: {None})
            eval_ {bool} -- If 'True', use 'eval' modede on net. (default: {True})
            grads {bool} -- If gradients should be computed (default: {False})
            to_cpu {bool} -- For larger data sets we need to move the results to cpu
                (default: {False})
            num_workers {int} -- Number of workes in created dataloader (default: {0})
            func {func} -- A toch function, such as `torch.sigmoid` which is called after the predict.
                (default: {None})
            **kwargs -- Passed to make_dataloader.

        Returns:
            [TupleTree, np.ndarray or tensor] -- Predictions
        """
        if not hasattr(self.net, "predict"):
            return self.predict_net(
                input,
                batch_size,
                numpy,
                eval_,
                grads,
                to_cpu,
                num_workers,
                is_dataloader,
                func,
                **kwargs,
            )

        pred_func = wrapfunc(func, self.net.predict)
        preds = self._predict_func(
            pred_func,
            input,
            batch_size,
            numpy,
            eval_,
            grads,
            to_cpu,
            num_workers,
            is_dataloader,
            **kwargs,
        )
        return array_or_tensor(preds, numpy, input)

    def save_model_weights(self, path, **kwargs):
        """Save the model weights.

        Parameters:
            path: The filepath of the model.
            **kwargs: Arguments passed to torch.save method.
        """
        return torch.save(self.net.state_dict(), path, **kwargs)

    def load_model_weights(self, path, **kwargs):
        """Load model weights.

        Parameters:
            path: The filepath of the model.
            **kwargs: Arguments passed to torch.load method.
        """
        self.net.load_state_dict(torch.load(path, **kwargs))

    def save_net(self, path, **kwargs):
        """Save self.net to file (e.g. net.pt).

        Arguments:
            path {str} -- Path to file.
            **kwargs are passed to torch.save

        Returns:
            None
        """
        return torch.save(self.net, path, **kwargs)

    def load_net(self, path, **kwargs):
        """Load net from file (e.g. net.pt), and set as self.net

        Arguments:
            path {str} -- Path to file.
            **kwargs are passed to torch.load

        Returns:
            torch.nn.Module -- self.net
        """
        self.net = torch.load(path, **kwargs)
        if hasattr(self, "device"):
            self.net.to(self.device)
        return self.net


def _get_element_in_dataloader(dataloader):
    dataset = dataloader.dataset
    try:
        return dataset[:2]
    except:
        pass
    try:
        return dataset[[0, 1]]
    except:
        pass
    try:
        return dataloader.collate_fn([dataset[0], dataset[1]])
    except:
        pass
    return None


def wrapfunc(outer, inner):
    """Essentially returns the function `lambda x: outer(inner(x))`
    If `outer` is None, return `inner`.
    """
    if outer is None:
        return inner

    def newfun(*args, **kwargs):
        return outer(inner(*args, **kwargs))

    return newfun



'''
Callbacks.
'''
import warnings
from collections import OrderedDict, defaultdict
from pathlib import Path
import math
import numpy as np
try:
    import pandas as pd
except:
    pass
import torch
import torchtuples
from . import lr_scheduler 
from torchtuples.utils import make_name_hash, TimeLogger


class Callback:
    '''Temple for how to write callbacks.
    '''
    def give_model(self, model):
        self.model = model

    def on_fit_start(self):
        pass

    def on_epoch_start(self):
        pass

    def on_batch_start(self):
        pass

    def before_step(self):
        """Called after loss.backward(), but before optim.step()."""
        pass

    def on_batch_end(self):
        pass

    def on_epoch_end(self):
        pass

    def on_fit_end(self):
        pass


class CallbackHandler:
    def __init__(self, callbacks):
        if type(callbacks) in (list, tuple, torchtuples.TupleTree):
            self.callbacks = OrderedDict()
            for c in callbacks:
                self[self._make_name(c)] = c
        else:
            self.callbacks = OrderedDict(callbacks)

    def _make_name(self, obj):
        name = type(obj).__name__
        i = 0
        new_name = name
        while new_name in self.keys():
            new_name = f"{name}_{i}"
            i += 1
            if i >= 100:
                raise RuntimeError("Stopped while loop. Too many callbacks with same name")
        return new_name
    
    def __getitem__(self, name):
        return self.callbacks[name]
    
    def __setitem__(self, name, callback):
        self.append(callback, name)

    def append(self, callback, name=None):
        if hasattr(self, 'model'):
            callback.give_model(self.model)
        if name is None:
            name = self._make_name(callback)
        assert name not in self.callbacks.keys(), f"Duplicate name: {name}"
        self.callbacks[name] = callback
    
    def items(self):
        return self.callbacks.items()
    
    def keys(self):
        return self.callbacks.keys()
    
    def values(self):
        return self.callbacks.values()
    
    def __len__(self):
        return len(self.callbacks)

    def apply_callbacks(self, func):
        stop_signal = False
        for c in self.values():
            stop = func(c)
            stop = stop if stop else False
            stop_signal = stop_signal or stop
        return stop_signal

    def give_model(self, model):
        self.model = model
        stop_signal = self.apply_callbacks(lambda x: x.give_model(model))
        return stop_signal

    def on_fit_start(self):
        stop_signal = self.apply_callbacks(lambda x: x.on_fit_start())
        return stop_signal

    def on_epoch_start(self):
        stop_signal = self.apply_callbacks(lambda x: x.on_epoch_start())
        return stop_signal

    def on_batch_start(self):
        stop_signal = self.apply_callbacks(lambda x: x.on_batch_start())
        return stop_signal

    def before_step(self):
        stop_signal = self.apply_callbacks(lambda x: x.before_step())
        return stop_signal

    def on_batch_end(self):
        stop_signal = self.apply_callbacks(lambda x: x.on_batch_end())
        return stop_signal

    def on_epoch_end(self):
        stop_signal = self.apply_callbacks(lambda x: x.on_epoch_end())
        return stop_signal

    def on_fit_end(self):
        stop_signal = self.apply_callbacks(lambda x: x.on_fit_end())
        return stop_signal


class TrainingCallbackHandler(CallbackHandler):
    '''Object for holding all callbacks.
    '''
    def __init__(self, optimizer, train_metrics, log, val_metrics=None, callbacks=None):
        super().__init__(dict(log=log, optimizer=optimizer, train_metrics=train_metrics))
        if val_metrics:
            self['val_metrics'] = val_metrics
        if callbacks is not None:
            if type(callbacks) in (list, tuple, torchtuples.TupleTree):
                for c in callbacks:
                    self.append(c)
            else:
                for name, c in callbacks.items():
                    self[name] = c

    def append(self, callback, name=None):
        super().append(callback, name)
        self.callbacks.move_to_end('log')


class TrainingLogger(Callback):
    def __init__(self, verbose=1):
        self.epoch = 0
        self.epochs = []
        self.loss = []
        self._verbose = verbose
    
    @property
    def monitors(self):
        return self._monitors

    @monitors.setter
    def monitors(self, monitor_dict):
        self._monitors = monitor_dict

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = value

    def on_fit_start(self):
        self.time_logger = TimeLogger()

    def on_epoch_end(self):
        self.epochs.append(self.epoch)
        if self.verbose:
            self.print_on_epoch_end()
        self.epoch += 1
        return False
    
    def print_on_epoch_end(self):
        tot, prev = self.time_logger.hms_diff()
        string = f"{self.epoch}:\t[{prev} / {tot}],\t"
        print(string + self.get_measures())

    def get_measures(self):
        measures = self.monitors
        if self.verbose.__class__ in [dict, OrderedDict]:
            measures = OrderedDict(measures, **self.verbose)
        string = ''
        for prefix, mm in measures.items():
            for name, score in mm.to_pandas().iloc[-1].items(): # slow but might not matter
                if (score is not None) and (np.isnan(score) == False):
                    string += '\t%s:' % (prefix + name)
                    string += ' %.4f,' % score
        return string[:-1]

    def to_pandas(self, colnames=None):
        '''Get data in dataframe.
        '''
        assert colnames is None, 'Not implemented'
        dfs = []
        for prefix, mm in self.monitors.items():
            df = mm.to_pandas()
            df.columns = prefix + df.columns
            dfs.append(df)
        df = pd.concat(dfs, axis=1)
        return df

    def plot(self, colnames=None, **kwargs):
        return self.to_pandas(colnames).plot(**kwargs)



class MonitorMetrics(Callback):
    def __init__(self, per_epoch=1):
        self.scores = dict()
        self.epoch = -1
        self.per_epoch = per_epoch

    def on_epoch_end(self):
        self.epoch += 1

    def append_score_if_epoch(self, name, val):
        if self.epoch % self.per_epoch == 0:
            self.append_score(name, val)
    
    def append_score(self, name, val):
        scores = self.scores.get(name, {'epoch': [], 'score': []})
        scores['epoch'].append(self.epoch)
        scores['score'].append(val)
        self.scores[name] = scores

    def to_pandas(self):
        '''Return scores as a pandas dataframe'''
        scores = [pd.Series(score['score'], index=score['epoch']).rename(name)
                  for name, score in self.scores.items()]
        scores = pd.concat(scores, axis=1)
        if type(scores) is pd.Series:
            scores = scores.to_frame()
        return scores


class _MonitorFitMetricsTrainData(MonitorMetrics):
    # def __init__(self, per_epoch=1):
    #     super().__init__(per_epo)
        # self.per_epoch = per_epoch

    def on_epoch_start(self):
        self.batch_metrics = defaultdict(list)
        return super().on_epoch_start()
    
    def on_batch_end(self):
        for name, score in self.model.batch_metrics.items():
            self.batch_metrics[name].append(score.item())
        return super().on_batch_end()

    def on_epoch_end(self):
        super().on_epoch_end()
        for name, vals in self.batch_metrics.items():
            self.append_score_if_epoch(name, np.mean(vals))
        # if self.epoch % self.per_epoch != 0:
        #     return False
        # for name, vals in self.batch_metrics.items():
        #     self.append_score(name, np.mean(vals))
        # return False


class MonitorFitMetrics(MonitorMetrics):
    def __init__(self, dataloader=None, per_epoch=1):
        super().__init__(per_epoch)
        self.dataloader = dataloader

    def on_epoch_end(self):
        super().on_epoch_end()
        if self.epoch % self.per_epoch != 0:
            return None
        if self.dataloader is None:
            scores = {name: np.nan for name in self.model.metrics.keys()}
        else:
            scores = self.model.score_in_batches_dataloader(self.dataloader)
        for name, val in scores.items():
            self.append_score(name, val)


class MonitorTrainMetrics(Callback):
    '''Monitor metrics for training loss.

    Parameters:
        per_epoch: How often to calculate.
    '''
    @property
    def scores(self):
        return self.model.train_metrics.scores
    
    def to_pandas(self):
        '''Return scores as a pandas dataframe'''
        scores = [pd.Series(score[name], index=score['epoch']).rename(name)
                  for name, score in self.scores.items()]
        scores = pd.concat(scores, axis=1)
        if type(scores) is pd.Series:
            scores = scores.to_frame()
        return scores


# class MonitorXy(MonitorBase):
#     '''Monitor metrics for classification and regression.
#     Same as MonitorBase but we input a pair, X, y instead of data.

#     For survival methods, see e.g. MonitorCox.

#     Parameters:
#         X: Numpy array with features.
#         y: Numpy array with labels.
#         monitor_funcs: Function, list, or dict of functions giving quiatities that should
#             be monitored.
#             The function takes argumess (df, preds) and should return a score.
#         batch_size: Batch size used for calculating the scores.
#         **kwargs: Can be passed to predict method.
#     '''
#     def __init__(self, X, y, monitor_funcs, per_epoch=1, batch_size=512, **kwargs):
#         self.X, self.y = X, y
#         self.kwargs = kwargs
#         super().__init__(monitor_funcs, per_epoch, batch_size)

#     def get_score_inputs(self):
#         '''This function should create arguments to the monitor function.
#         Typically it can return a tuple with (y_true, preds), to calculate e.g. auc.
#         '''
#         preds = self.model.predict(self.X, self.batch_size, **self.kwargs)
#         return self.y, preds


# class MonitorSklearn(MonitorXy):
#     '''Class for monitoring metrics of from sklearn metrics

#     Parameters:
#         X: Numpy array with features.
#         y: Numpy array with labels.
#         monitor: Name for method in sklearn.metrics.
#             E.g. 'log_loss', or pass sklearn.metrics.log_loss.
#             For additional parameter, specify the function with a lambda statemetn.
#                 e.g. {'accuracy': lambda y, p: metrics.accuracy_score(y, p > 0.5)}
#         batch_size: Batch size used for calculating the scores.
#         **kwargs: Can be passed to predict method.
#     '''
#     def __init__(self, X, y, monitor, per_epoch=1, batch_size=512, **kwargs):
#         if monitor.__class__ is str:
#             monitor = {monitor: monitor}
#         elif monitor.__class__ is list:
#             monitor = {mon if mon.__class__ is str else str(i): mon
#                        for i, mon in enumerate(monitor)}

#         monitor = {name: getattr(metrics, mon) if mon.__class__ == str else mon
#                    for name, mon in monitor.items()}
#         super().__init__(X, y, monitor, per_epoch, batch_size)


class ClipGradNorm(Callback):
    '''Callback for clipping gradients.
    
    See torch.nn.utils.clip_grad_norm_.

    Parameters:
        net: Network wtih parameters() function.
        max_norm (float or int): max norm of the gradients
        norm_type (float or int): type of the used p-norm. Can be ``'inf'`` for infinity norm.

    '''
    def __init__(self, net, max_norm, norm_type=2):
        self.net = net
        self.max_norm = max_norm
        self.norm_type = norm_type

    def before_step(self):
        torch.nn.utils.clip_grad_norm_(self.net.parameters(), self.max_norm, self.norm_type)
        stop_signal = False
        return stop_signal


class LRScheduler(Callback):
    '''Wrapper for pytorch.optim.lr_scheduler objects.

    Parameters:
        scheduler: A pytorch.optim.lr_scheduler object.
        mm_obj: Monitor object, where first metric is used for early stopping.
            E.g. MonitorSurvival(df_val, 'cindex').
    '''
    def __init__(self, scheduler, mm_obj):
        self.scheduler = scheduler
        self.mm_obj = mm_obj

    def on_epoch_end(self):
        score = self.mm_obj.scores[0][-1]
        self.scheduler.step(score)
        stop_signal = False
        return stop_signal


class LRSchedulerBatch(Callback):
    '''Wrapper for schedulers

    Parameters:
        scheduler: A scheduler, e.g. BatchCosineAnnealingLR()
    '''
    def __init__(self, scheduler):
        self.scheduler = scheduler

    def on_fit_start(self):
        self.scheduler.step()  
    
    def on_batch_end(self):
        self.scheduler.step()
        return False


class LRCosineAnnealing(LRSchedulerBatch):
    def __init__(self, cycle_len=1, cycle_multiplier=2, cycle_eta_multiplier=1., eta_min=0):
        self.first_cycle_len = cycle_len
        self.cycle_multiplier = cycle_multiplier
        self.cycle_eta_multiplier = cycle_eta_multiplier
        self.eta_min = eta_min
        scheduler = None
        super().__init__(scheduler)
    
    def on_fit_start(self):
        # if self.first_cycle_len == "epoch":
        if self.scheduler is None:
            cycle_len = self.first_cycle_len * self.model.fit_info['batches_per_epoch']
            scheduler = lr_scheduler.LRBatchCosineAnnealing(self.model.optimizer, int(cycle_len),
                                                            self.cycle_multiplier,
                                                            self.cycle_eta_multiplier,
                                                            self.eta_min,
                                                            keep_etas=True)
            self.scheduler = scheduler
        elif self.model.optimizer is not self.scheduler.optimizer:
            raise RuntimeError(
                "Changed optimizer, and we have not implemented cosine annealing for this")
        super().on_fit_start()
    
    def get_cycle_len(self):
        return self.scheduler.cycle_len
    
    def get_epochs_per_cycle(self):
        return self.get_cycle_len() / self.model.fit_info['batches_per_epoch']

    def get_etas(self):
        return self.scheduler.etas

    def to_pandas(self):
        return self.scheduler.to_pandas()

    def plot(self, **kwargs):
        return self.scheduler.plot(**kwargs)


class LRFinder(Callback):
    def __init__(self, lr_lower=1e-7, lr_upper=10., n_steps=100, tolerance=np.inf):
        self.lr_lower = lr_lower
        self.lr_upper = lr_upper
        self.n_steps = n_steps
        self.lowest_loss = np.inf
        self.tolerance = tolerance

    def on_fit_start(self):
        self.batch_loss = []
        self.scheduler = lr_scheduler.LRFinderScheduler(self.model.optimizer, self.lr_lower,
                                                        self.lr_upper, self.n_steps)
        # self.scheduler.step()

    def on_batch_start(self):
        self.scheduler.step()
        for group in self.model.optimizer.param_groups:
            group['initial_lr'] = group['lr']  # Needed when using weight decay.
    
    def on_batch_end(self):
        # self.scheduler.step()
        batch_loss = self.model.batch_loss.item()
        self.batch_loss.append(batch_loss)
        if batch_loss > self.tolerance:
            return True
        # if (batch_loss / self.lowest_loss) > self.tolerance:
        #     return True
        self.lowest_loss = min(self.lowest_loss, batch_loss)
        if self.scheduler.batch_iter == self.n_steps:
            return True
        return False
    
    def to_pandas(self, smoothed=0):
        res = pd.DataFrame(dict(train_loss=self.batch_loss),
                           index=self.lrs[:len(self.batch_loss)])
        res.index.name = 'lr'
        if smoothed:
            res = res.apply(_smooth_curve, beta=smoothed)
        return res
    
    def plot(self, logx=True, smoothed=0.98, **kwargs):
        res = self.to_pandas(smoothed)
        ylabel = 'bach_loss'
        if smoothed:
            ylabel = ylabel + ' (smoothed)'
        ax = res.plot(logx=logx, **kwargs)
        ax.set_xlabel('lr')
        ax.set_ylabel(ylabel)
        return ax

    @property
    def lrs(self):
        return self.scheduler.lrs
    
    def get_best_lr(self, lr_min=1e-4, lr_max=1., _multiplier=10.):
        """Get suggestion for bets learning rate.
        It is beter to investigate the plot, but this might work too.
        
        Keyword Arguments:
            lower {float} -- Lower accepable learning rate (default: {1e-4})
            upper {float} -- Upper acceptable learning rate (default: {1.})
            _multiplier {float} -- See sorce code (default according to fast.ai) (default: {10})
        
        Returns:
            float -- Suggested best learning rate.
        """
        smoothed = _smooth_curve(self.batch_loss)
        idx_min = np.argmin(smoothed)
        best_lr = self.lrs[idx_min] / _multiplier
        best_lr = np.clip(best_lr, lr_min, lr_max)
        return best_lr

        # idx_lower = np.searchsorted(self.lrs, lr_min * _multiplier)
        # idx_upper = np.searchsorted(self.lrs, lr_max * _multiplier, 'right')
        # smoothed = _smooth_curve(self.batch_loss)[idx_lower:idx_upper]
        # idx_min = np.argmin(smoothed)
        # best_lr = self.lrs[idx_lower:idx_upper][idx_min] / _multiplier
        # return best_lr

def _smooth_curve(vals, beta=0.98):
    """From fastai"""
    avg_val = 0
    smoothed = []
    for (i,v) in enumerate(vals):
        avg_val = beta * avg_val + (1-beta) * v
        smoothed.append(avg_val/(1-beta**(i+1)))
    return smoothed


class DecoupledWeightDecay(Callback):
    """Same weight decay for all groups in the optimizer."""
    def __init__(self, weight_decay, normalized=False, nb_epochs=None):
        if (weight_decay >= 1) or (weight_decay < 0):
            warnings.warn(f"Weigth decay should be in in [0, 1), got {weight_decay}")
        self.weight_decay = weight_decay
        self.normalized = normalized
        # if self.normalized:
        #     if not ((type(nb_epochs) is int) or callable(nb_epochs)):
        #         raise ValueError(f"Need nb_epochs to be callable or int, not {nb_epochs}")
        self.nb_epochs = nb_epochs
    
    def on_fit_start(self):
        self._batches_per_epoch = self.model.fit_info['batches_per_epoch']

    @property
    def nb_epochs(self):
        if type(self._nb_epochs) is int:
            return self._nb_epochs
        elif callable(self._nb_epochs):
            return self._nb_epochs()
        raise RuntimeError('nb_epochs needs to be callable or int')

    @nb_epochs.setter
    def nb_epochs(self, nb_epochs):
        if self.normalized:
            if not ((type(nb_epochs) is int) or callable(nb_epochs)):
                raise ValueError(f"Need nb_epochs to be callable or int, not {type(nb_epochs)}")
        self._nb_epochs = nb_epochs
    
    def _normalized_weight_decay(self):
        # if type(self.nb_epochs) is int:
        #     nb_epochs = self.nb_epochs
        # elif callable(self.nb_epochs):
        #     nb_epochs = self.nb_epochs()
        # else:
        #     raise RuntimeError('nb_epochs needs to be callable or int')
        norm_const = math.sqrt(1 / (self._batches_per_epoch * self.nb_epochs))
        return self.weight_decay * norm_const
    
    def before_step(self):
        # Weight decay out of the loss. After the gradient computation but before the step.
        if self.normalized:
            weight_decay = self._normalized_weight_decay()
        else:
            weight_decay = self.weight_decay
        for group in self.model.optimizer.param_groups:
            lr = group['lr']
            # alpha = group.get('initial_lr', 1.)
            alpha = group.get('initial_lr', lr)
            eta = lr / alpha
            for p in group['params']:
                if p.grad is not None:
                    p.data = p.data.add(-weight_decay * eta, p.data)
                    # p.data.mul_(1 - weight_decay * eta)
        return False


# class EarlyStoppingCycle_old(Callback):
#     '''
#     TODO: Should rewrite with _ActionOnBestMetric.

#     Stop training when monitored quantity has not improved the last cycle.
#     Takes a Monitor object that is also a callback.
#     Use first metric in mm_obj to determine early stopping.

#     Parameters:
#         mm_obj: Monitor object, where first metric is used for early stopping.
#             E.g. MonitorSurvival(df_val, 'cindex').
#         get_score: Function for obtaining current scores. If string, we use validation metric.
#             Default is 'loss' which gives validation loss.
#         minimize: If we are to minimize or maximize monitor.
#         min_delta: Minimum change in the monitored quantity to qualify as an improvement,
#             i.e. an absolute change of less than min_delta, will count as no improvement.
#         patience: Number of cycles patience.
#         model_file_path: If spesified, the model weights will be stored whever a better score
#             is achieved.
#     '''
#     def __init__(self, lr_scheduler='optimizer', get_score='loss', minimize=True, min_delta=0,
#                  patience=1, min_cycles=4, model_file_path=None, load_best=False):
#         self.get_score = get_score
#         self._get_score = self.get_score
#         self.minimize = minimize
#         self.min_delta = min_delta
#         self.patience = patience
#         self.min_cycles = min_cycles
#         self.model_file_path = model_file_path
#         self.load_best = load_best
#         if self.load_best and (self.model_file_path is None):
#             raise ValueError("To use 'load_best' you need to provide a model_file_path")
#         self.lr_scheduler = lr_scheduler
#         self.cur_best = np.inf if self.minimize else -np.inf
#         self.cur_best_cycle_nb = None

#     def on_fit_start(self):
#         if type(self.get_score) is str:
#             self.get_score = lambda: self.model.val_metrics.scores[self._get_score]['score'][-1]
#         if self.lr_scheduler == 'optimizer':
#             self.lr_scheduler = self.model.optimizer.lr_scheduler
        
#     def on_epoch_end(self):
#         etas = self.lr_scheduler.get_etas()[:-1]  # Drop last because allready updated by optimizer
#         cycle_nb = (np.diff(etas) > 0).sum()
#         score = self.get_score()

#         if self.minimize:
#             if score < (self.cur_best - self.min_delta):
#                 self.cur_best = score
#                 self.cur_best_cycle_nb = cycle_nb
#         else:
#             if score > (self.cur_best + self.min_delta):
#                 self.cur_best = score
#                 self.cur_best_cycle_nb = cycle_nb

#         if (score == self.cur_best) and (self.model_file_path is not None):
#             self.model.save_model_weights(self.model_file_path)

#         stop_signal = ((cycle_nb > (self.cur_best_cycle_nb + self.patience)) and 
#                        (cycle_nb >= self.min_cycles))
#         return stop_signal

#     def on_fit_end(self):
#         if self.load_best:
#             self.model.load_model_weights(self.model_file_path)
#         return super().on_fit_end()


class StopIfExplodeOrNan(Callback):
    """Stop trainig if training or validation loss becomes larger than a threshold or becomes nan.
    Checks both train and val data.

    Keyword Arguments:
        threshold {float} -- Stop if train or val loss is 'nan' or larger than threshold (default: {np.inf})
        metric {str} -- Whick metric in model.log.monitors should be used.(default: {'loss'})
    """
    def __init__(self, threshold=np.inf, metric='loss'):
        self.threshold = threshold
        self.metric = metric

    def _update_cur_best(self, key):
        score = self.model.log.monitors[key].scores[self.metric]['score'][-1]
        if np.isnan(score):
            return True
        if score >= self.threshold:
            return True
        return False

    def on_epoch_end(self):
        return self._update_cur_best('train_') or self._update_cur_best('val_')


class _ActionOnBestMetric(Callback):
    """Abstract class used for e.g. EarlyStopping.
    """
    def __init__(self, metric='loss', dataset='val', get_score=None, minimize=True, min_delta=0.,
                 checkpoint_model=True, file_path=None, load_best=True, rm_file=True):
        self.metric = metric
        self.dataset = dataset
        self.get_score = get_score
        self.minimize = minimize
        self.min_delta = min_delta
        self._checkpoint_model = checkpoint_model
        if not self._checkpoint_model:
            assert load_best == False, "Need load best to be False when '_checkpoint_model' is False"
        if (not load_best and rm_file) and self._checkpoint_model:
            raise ValueError("If you really want not not load best but remove file you can instead remove this callback.")
        self.load_best = load_best
        self.rm_file = rm_file
        self.file_path = file_path if file_path else make_name_hash('weight_checkpoint')
        self.file_path = Path(self.file_path)
        self.cur_best = np.inf if self.minimize else -np.inf
        self._iter_since_best = 0

    def on_fit_start(self):
        if not self.file_path.exists() and self._checkpoint_model:
            self.model.save_model_weights(self.file_path)
        if self.get_score is None:
            if self.dataset == 'val':
                metrics = self.model.val_metrics
            elif self.dataset == 'train':
                metrics = self.model.train_metrics
            else:
                raise ValueError("Need dataset to be 'val' or 'train'.")
            self.get_score = lambda: metrics.scores[self.metric]['score'][-1]

    def on_epoch_end(self):
        score = self.get_score()
        if self.minimize:
            if score < (self.cur_best - self.min_delta):
                self.cur_best = score
                self._iter_since_best = -1
        else:
            if score > (self.cur_best + self.min_delta):
                self.cur_best = score
                self._iter_since_best = -1
        
        if (score == self.cur_best) and (self._checkpoint_model):
            self.model.save_model_weights(self.file_path)
        self._iter_since_best += 1

    def on_fit_end(self):
        if self._checkpoint_model:
            if self.load_best:
                self.load_weights()
            if self.rm_file:
                self.rm_weight_file()

    def load_weights(self):
        if not self._checkpoint_model:
            raise RuntimeError("This model has no stored weights")
        self.model.load_model_weights(self.file_path)

    def rm_weight_file(self):
        if self.file_path.exists():
            self.file_path.unlink()


class BestWeights(_ActionOnBestMetric):
    """
    Keep trac of the weight of the best performin model.
    If you also want early stopping, you can use EarlyStopping or EarlyStoppingCycle instead.
    
    Keyword Arguments:
        metric {str} -- Metric in model.train_metrics or model.val_metrics (default: {'loss'})
        dataset {str} -- Data set which is moitored train/val (default: {'val'})
        get_score {[type]} -- Alternative to metric, where you can give a function that returns the
            scores. (default: {None})
        minimize {bool} -- If we are minimizing or maximizing the score (default: {True})
        file_path {[type]} -- Alternative file path for model weight. If 'None' we generate one.
            (default: {None})
        load_best {bool} -- Load best weight into model object after training.
            If 'False' this needs to be done by calling the method 'load_weights' (default: {True})
        rm_file {bool} -- If we should delete the checkpoint weight file after finishing training.
            (default: {True})
    """
    def __init__(self, metric='loss', dataset='val', get_score=None, minimize=True, file_path=None,
                 load_best=True, rm_file=True):
        min_delta = 0.
        checkpoint_model = True
        super().__init__(metric, dataset, get_score, minimize, min_delta, checkpoint_model,
                         file_path, load_best, rm_file)


class EarlyStopping(_ActionOnBestMetric):
    """
    Stop training when monitored quantity has not improved the last epochs.
    
    Keyword Arguments:
        metric {str} -- Metric in model.train_metrics or model.val_metrics (default: {'loss'})
        dataset {str} -- Data set which is moitored train/val (default: {'val'})
        get_score {[type]} -- Alternative to metric, where you can give a function that returns the
            scores. (default: {None})
        minimize {bool} -- If we are minimizing or maximizing the score (default: {True})
        min_delta {[type]} -- Improvement required to consider the new score better than the
            previous best. (default: {0.})
        patience {int} -- Number of epochs to wait since the best score before stopping. (default: {10})
        checkpoint_model {bool} -- If we should keep track of the best model weights. (default: {True})
        file_path {[type]} -- Alternative file path for model weight. If 'None' we generate one.
            (default: {None})
        load_best {bool} -- Load best weight into model object after training.
            If 'False' this needs to be done by calling the method 'load_weights' (default: {True})
        rm_file {bool} -- If we should delete the checkpoint weight file after finishing training.
            (default: {True})
    """
    def __init__(self, metric='loss', dataset='val', get_score=None, minimize=True, min_delta=0.,
                 patience=10, checkpoint_model=True, file_path=None, load_best=True, rm_file=True):
        self.patience = patience
        super().__init__(metric, dataset, get_score, minimize, min_delta, checkpoint_model,
                         file_path, load_best, rm_file)

    def on_epoch_end(self):
        super().on_epoch_end()
        return self._iter_since_best >= self.patience


class EarlyStoppingCycle(_ActionOnBestMetric):
    """
    Stop training when monitored quantity has not improved the last cycles.
    
    Keyword Arguments:
        metric {str} -- Metric in model.train_metrics or model.val_metrics (default: {'loss'})
        dataset {str} -- Data set which is moitored train/val (default: {'val'})
        lr_scheduler {str} -- lr_scheduler object. If 'optimizer' use model.optimizer.lr_scheduler.
            (default: {'optimizer'})
        get_score {[type]} -- Alternative to metric, where you can give a function that returns the
            scores. (default: {None})
        minimize {bool} -- If we are minimizing or maximizing the score (default: {True})
        min_delta {[type]} -- Improvement required to consider the new score better than the
            previous best. (default: {0.})
        patience {int} -- Number of cycles to wait since the best score before stopping. (default: {1})
        min_cycles {int} -- Minimum number of cycles required before stopping. (default: {4})
        checkpoint_model {bool} -- If we should keep track of the best model weights. (default: {True})
        file_path {[type]} -- Alternative file path for model weight. If 'None' we generate one.
            (default: {None})
        load_best {bool} -- Load best weight into model object after training.
            If 'False' this needs to be done by calling the method 'load_weights' (default: {True})
        rm_file {bool} -- If we should delete the checkpoint weight file after finishing training.
            (default: {True})
    """
    def __init__(self, metric='loss', dataset='val', lr_scheduler='optimizer', get_score=None,
                 minimize=True, min_delta=0., patience=1, min_cycles=4, checkpoint_model=True,
                 file_path=None, load_best=True, rm_file=True):
        self.lr_scheduler = lr_scheduler
        self.patience = patience
        self.min_cycles = min_cycles
        self.cur_best_cycle_nb = 0
        super().__init__(metric, dataset, get_score, minimize, min_delta, checkpoint_model,
                         file_path, load_best, rm_file)

    def on_fit_start(self):
        if self.lr_scheduler == 'optimizer':
            self.lr_scheduler = self.model.optimizer.lr_scheduler
        return super().on_fit_start()

    def get_current_cycle_nb(self):
        """Get current cycle number"""
        etas = self.lr_scheduler.get_etas()[:-1]  # Drop last because allready updated by optimizer
        return (np.diff(etas) > 0).sum()

    def on_epoch_end(self):
        cycle_nb = self.get_current_cycle_nb()
        score = self.get_score()

        if self.minimize:
            if score < (self.cur_best - self.min_delta):
                self.cur_best = score
                self._iter_since_best = -1
                self.cur_best_cycle_nb = cycle_nb
        else:
            if score > (self.cur_best + self.min_delta):
                self.cur_best = score
                self._iter_since_best = -1
                self.cur_best_cycle_nb = cycle_nb
        self._iter_since_best += 1
        
        if (score == self.cur_best) and (self._checkpoint_model):
            self.model.save_model_weights(self.file_path)

        stop_signal = ((cycle_nb > (self.cur_best_cycle_nb + self.patience)) and 
                       (cycle_nb >= self.min_cycles))
        return stop_signal




import functools
import itertools
from typing import Any, Union

import numpy as np
import torch

import torchtuples


def apply_leaf(func):
    """Apply a function to data in TupleTree objects (leaf nodes).

    E.g.: Two ways to get shapes of all elements in a tuple

    data = tuplefy([(torch.randn(1, 2), torch.randn(2, 3)),
                    torch.randn(4, 4)])

    # Method 1:
    @apply_leaf
    def shape_of(data):
        return len(data)

    shape_of(data)

    # Method 2:
    apply_leaf(lambda x: x.shape)(data)
    """

    @functools.wraps(func)
    def wrapper(data, *args, **kwargs):
        if type(data) in _CONTAINERS:
            return TupleTree(wrapper(sub, *args, **kwargs) for sub in data)
        return func(data, *args, **kwargs)

    return wrapper


def reduce_leaf(func, init_func=None):
    """Reduce operation on TupleTree objects.
    It reduces the leaf nodes to the first TupleTree's topology.

    Exs:
    a = ((1, (2, 3), 4),
         (1, (2, 3), 4),
         (1, (2, 3), 4),)
    a = tuplefy(a)
    reduce_leaf(lambda x, y: x+y)(a)

    Gives:
    (3, (6, 9), 12)
    """

    def reduce_rec(acc_val, val, **kwargs):
        if type(acc_val) in _CONTAINERS:
            return TupleTree(reduce_rec(av, v) for av, v in zip(acc_val, val))
        return func(acc_val, val, **kwargs)

    @functools.wraps(func)
    def wrapper(data, **kwargs):
        if not data.to_levels().all_equal():
            raise ValueError(
                "Topology is not the same for all elements in data, and can not be reduced"
            )
        iterable = iter(data)
        if init_func is None:
            acc_val = next(iterable)
        else:
            acc_val = data[0].apply(init_func)
        for val in iterable:
            acc_val = reduce_rec(acc_val, val, **kwargs)
        return acc_val

    return wrapper


def all_equal(data):
    """All typles (from top level) are the same
    E.g. (a, a, a)
    """
    return data.apply_nrec(lambda x: x == data[0]).all()


def get_if_all_equal(data, default=None):
    """Get value of all are the same, else return default value.

    Arguments:
        data {TupleTree} -- TupleTree data.

    Keyword Arguments:
        default {any} -- Return if all are not equal (default: {None})
    """
    if data.all_equal():
        return data[0]
    return default


def val_if_single(data):
    """Get value if there is only one element in the tupletree.

    Arguments:
        data {TupleTree} -- TupleTree data.
    """
    if len(TupleTree(data).flatten()) == 1:
        return data[0]
    return data


def zip_leaf(data):
    """Aggregate data to a list of the data.
    This is essentialy a zip opperation that works on the leaf nodes
    ((a1, (a2, a3)), (b1, (b2, b3))) -> ([a1, b1], ([a2, b2], [a3, b3]))

    Inverse of unzip_leaf
    """
    init_func = lambda _: list()

    def append_func(list_, val):
        list_.append(val)
        return list_

    return reduce_leaf(append_func, init_func)(data)


@apply_leaf
def shapes_of(data):
    """Apply x.shape to elemnts in data."""
    return data.shape


@apply_leaf
def lens_of(data):
    """Apply len(x) to elemnts in data."""
    return len(data)


@apply_leaf
def dtypes_of(data):
    """Apply x.dtype to elemnts in data."""
    return data.dtype


@apply_leaf
def numpy_to_tensor(data):
    """Transform numpy arrays to torch tensors."""
    return torch.from_numpy(data)


@apply_leaf
def tensor_to_numpy(data):
    """Transform torch tensort arrays to numpy arrays."""
    if hasattr(data, "detach"):
        data = data.detach()
    if type(data) is torch.Size:
        return np.array(data)
    return data.cpu().numpy()


@apply_leaf
def astype(data, dtype, *args, **kwargs):
    """Change type to dtype.

    torch tensors: we call 'data.type(dtype, *args, **kwargs)'
    numpy arrays: we call 'data.astype(dtype, *args, **kwargs)'
    """
    if type(data) is torch.Tensor:
        return data.type(dtype, *args, **kwargs)
    elif type(data) is np.ndarray:
        return data.astype(dtype, *args, **kwargs)
    else:
        return RuntimeError(
            f"""Need 'data' to be torch.tensor of np.ndarray, found {type(data)}.
            """
        )


@apply_leaf
def types_of(data):
    """Returns all types in data"""
    return type(data)


def type_of(data):
    """Returns THE type of subelements in data.
    Hence, all elements in data needs to have the same class.
    """
    types = data.types().flatten()
    if types.count(types[0]) != len(types):
        raise ValueError("All objects in 'data' doest have the same type.")
    return types[0]


def is_flat(data):
    """Returns true if the TupleTree data is flat"""
    if type(data) not in _CONTAINERS:
        return True
    return all(data.apply_nrec(lambda x: type(x) not in _CONTAINERS))


def flatten_tuple(data):
    """Flatten the TupleTree data"""
    if type(data) not in _CONTAINERS:
        return data
    new = TupleTree(sub if type(sub) in _CONTAINERS else (sub,) for sub in data)
    new = TupleTree(itertools.chain.from_iterable(new))
    if new.is_flat():
        return new
    return flatten_tuple(new)


def tuple_levels(data, level=-1):
    """Replaces objects with the level they are on.

    Arguments:
        data {list or tuple} -- Data

    Keyword Arguments:
        level {int} -- Start level. Default of -1 gives flat list levels 0 (default: {-1})

    Returns:
        tuple -- Levels of objects
    """
    if type(data) not in _CONTAINERS:
        return level
    return TupleTree(tuple_levels(sub, level + 1) for sub in data)


def cat(seq, dim=0):
    """Conatenate tensors/arrays in tuple.
    Only works for dim=0, meaning we concatenate in the batch dim.
    """
    if not seq.shapes().apply(lambda x: x[1:]).all_equal():
        raise ValueError("Shapes of merged arrays need to be the same")

    type_ = seq.type()
    agg = seq.zip_leaf()
    if type_ is torch.Tensor:
        return agg.apply(lambda x: torch.cat(x, dim=dim))
    elif type_ is np.ndarray:
        return agg.apply(lambda x: np.concatenate(x, axis=dim))
    raise RuntimeError(f"Need type to be np.ndarray or torch.Tensor, fournd {type_}.")


def stack(seq, dim=0):
    """Stack tensors in tuple, see torch.stack.
    Only works for dim=0.
    """
    if dim != 0:
        raise NotImplementedError
    if not seq.shapes().all_equal():
        raise ValueError("Shapes of merged arrays need to be the same")
    type_ = seq.type()
    agg = seq.zip_leaf()
    if type_ is torch.Tensor:
        return agg.apply(torch.stack)
    elif type_ is np.ndarray:
        raise NotImplementedError
    raise RuntimeError(f"Need type to be np.ndarray or torch.Tensor, fournd {type_}.")


def split(data, split_size, dim=0):
    """Use torch.split and create multiple TupleTree with the splitted tensors."""
    if dim != 0:
        raise NotImplementedError
    if data.type() is not torch.Tensor:
        raise NotImplementedError(
            "Only implemented for torch tensors because np.split works differently"
        )

    splitted = data.apply(lambda x: x.split(split_size))
    return unzip_leaf(splitted)


def unzip_leaf(agg):
    """The inverse opeation of zip_leaf.
    This is essentialy a zip(*agg) opperation that works on the leaf nodes
    ([a1, b1], ([a2, b2], [a3, b3])) -> ((a1, (a2, a3)), (b1, (b2, b3)))
    """
    if type(agg) in _CONTAINERS:
        new = agg.apply_nrec(unzip_leaf)
        return TupleTree(zip(*new)).tuplefy()
    return agg


def tuplefy(*data, types=(list, tuple), stop_at_tuple=True):
    """Make TupleTree object from *args.

    Keyword Arguments:
        types {tuple} -- Types that should be transformed to TupleTree (default: {(list, tuple)})
        stop_at_tuple {bool} -- If 'True', the recusion stops at TupleTree elements,
            and if 'False' it will continue through TupleTree elements.

    Returns:
        TupleTree -- A TupleTree object
    """

    def _tuplefy(data, first=False):
        if (type(data) in types) or first:
            return TupleTree(_tuplefy(sub) for sub in data)
        return data

    if type(types) is type:
        types = [types]
    else:
        types = list(types)
    if not stop_at_tuple:
        types.extend(_CONTAINERS)
    if (len(data) == 1) and ((type(data[0]) in types) or (type(data[0]) in _CONTAINERS)):
        data = data[0]
    data = TupleTree(data)
    return _tuplefy(data, first=True)


@apply_leaf
def to_device(data: Any, device: Union[str, torch.device]) -> Any:
    """Move data to device

    Arguments:
        data {TupleTree, tensor} -- Tensors that should be moved to device.
        device {str, torch.device} -- Device data is moved to.

    Returns:
        TupleTree, tensor -- Data moved to device
    """
    return data.to(device)


def make_dataloader(
    data, batch_size, shuffle, num_workers=0, to_tensor=True, make_dataset=None, torch_ds_dl=False
):
    """Create a dataloder from tensor or np.arrays.

    Arguments:
        data {tuple, np.array, tensor} -- Data in dataloader e.g. (x, y)
        batch_size {int} -- Batch size used in dataloader
        shuffle {bool} -- If order should be suffled

    Keyword Arguments:
        num_workers {int} -- Number of workers in dataloader (default: {0})
        to_tensor {bool} -- Ensure that we use tensors (default: {True})
        make_dataset {callable} -- Function for making dataset. If 'None' we use
            DatasetTuple. (default {None}).
        torch_ds_dl {bool} -- If `True` we TensorDataset and DataLoader from torch. If
            `False` we use the (faster) versions from torchtuple (default {False}).

    Returns:
        DataLoaderBatch -- A dataloader object like the torch DataLoader
    """
    if to_tensor:
        data = tuplefy(data).to_tensor()
    if make_dataset is None:
        make_dataset = torchtuples.data.DatasetTuple
        if torch_ds_dl:
            make_dataset = torch.utils.data.TensorDataset
            # make_dataset = torchtuples.data.DatasetTupleSingle
    # dataset = DatasetTuple(data)
    dataset = make_dataset(*data)
    DataLoader = torchtuples.data.DataLoaderBatch
    if torch_ds_dl:
        DataLoader = torch.utils.data.DataLoader
    dataloader = DataLoader(dataset, batch_size, shuffle=shuffle, num_workers=num_workers)
    return dataloader


def docstring(doc_func):
    """Decorator to make function have the docstring of 'doc_func'."""

    def docstring_real(func):
        func.__doc__ = doc_func.__doc__
        return func

    return docstring_real


class TupleTree(tuple):
    """A tuple with some methods that works recursively.
    This is essentially a tree structure, with all internal nodes being TupleTree objects,
    and all leaf nodes contain data.

    Hence the apply methods is a map function on the leaf nodes.
    """

    @property
    def _constructor(self):
        return TupleTree

    def apply(self, func):
        """Shorthand to apply_leaf(func)(self)"""
        return apply_leaf(func)(self)

    def reduce(self, func, init_func=None, **kwargs):
        """Shorthand to reduce_leaf(func, init_func)(self).
        It reduces the leaf nodes to the first TupleTree's topology.

        Exs:
        a = ((1, (2, 3), 4),
             (1, (2, 3), 4),
             (1, (2, 3), 4),)
        tuplefy(a).reduce(lambda x, y: x+y)

        Gives: (3, (6, 9), 12)
        """
        return reduce_leaf(func, init_func)(self, **kwargs)

    def __add__(self, other):
        return self._constructor(super().__add__(other))

    @docstring(shapes_of)
    def shapes(self):
        return shapes_of(self)

    @docstring(lens_of)
    def lens(self):
        return lens_of(self)

    @docstring(dtypes_of)
    def dtypes(self):
        return dtypes_of(self)

    @docstring(numpy_to_tensor)
    def to_tensor(self):
        if self.type() is torch.Tensor:
            return self
        return numpy_to_tensor(self)

    @docstring(tensor_to_numpy)
    def to_numpy(self):
        if self.type() is np.ndarray:
            return self
        return tensor_to_numpy(self)

    @docstring(type_of)
    def type(self):
        return type_of(self)

    @docstring(types_of)
    def types(self):
        return types_of(self)

    @docstring(astype)
    def astype(self, dtype, *args, **kwargs):
        return astype(self, dtype, *args, **kwargs)

    @docstring(is_flat)
    def is_flat(self):
        return is_flat(self)

    @docstring(flatten_tuple)
    def flatten(self):
        return flatten_tuple(self)

    @docstring(tuple_levels)
    def to_levels(self):
        return tuple_levels(self)

    @property
    @docstring(tuple_levels)
    def levels(self):
        return tuple_levels(self)

    @docstring(cat)
    def cat(self, dim=0):
        return cat(self, dim)

    def reduce_nrec(self, func, initial=None):
        """Reduct non-recursive, only first list. See 'functools.reduce'."""
        if initial is None:
            return functools.reduce(func, self)
        else:
            return functools.reduce(func, self, initial)

    def apply_nrec(self, func):
        """Apply non-recursive, only first list"""
        return TupleTree(func(sub) for sub in self)

    def all(self):
        if not self.is_flat():
            raise RuntimeError("Need to have a flat structure to use 'all'")
        return all(self)

    @docstring(tuplefy)
    def tuplefy(self, types=(list, tuple), stop_at_tuple=False):
        self = tuplefy(self, types=types, stop_at_tuple=stop_at_tuple)
        return self

    @docstring(split)
    def split(self, split_size, dim=0):
        return split(self, split_size, dim)

    @docstring(zip_leaf)
    def zip_leaf(self):
        return zip_leaf(self)

    @docstring(unzip_leaf)
    def unzip_leaf(self):
        return unzip_leaf(self)

    @docstring(all_equal)
    def all_equal(self):
        """All typles (from top level) are the same
        E.g. (a, a, a)
        """
        return all_equal(self)

    @docstring(to_device)
    def to_device(self, device):
        return to_device(self, device)

    @docstring(make_dataloader)
    def make_dataloader(
        self,
        batch_size,
        shuffle,
        num_workers=0,
        to_tensor=True,
        make_dataset=None,
        torch_ds_dl=False,
    ):
        return make_dataloader(
            self, batch_size, shuffle, num_workers, to_tensor, make_dataset, torch_ds_dl
        )

    @property
    def iloc(self):
        """Used in as pd.DataFrame.iloc for subsetting tensors and arrays"""
        return _TupleTreeSlicer(self)

    @docstring(get_if_all_equal)
    def get_if_all_equal(self, default=None):
        return get_if_all_equal(self, default)

    @docstring(stack)
    def stack(self, dim=0):
        return stack(self, dim)

    def repeat(self, times):
        return self._constructor((self,) * times)

    def pipe(self, func, *args, **kwargs):
        """Shorthand for func(self, *args, **kwargs)"""
        return func(self, *args, **kwargs)

    @docstring(val_if_single)
    def val_if_single(self):
        return val_if_single(self)

    def numerate(self):
        """Replace leaf nodes with numbers from 0 to num. leaf nodes.
        Gives call order of apply function.

        Returns:
            TupleTree -- Numerated leaf nodes.
        """
        counter = itertools.count()
        return self.apply(lambda x: next(counter))

    def reorder(self, order):
        """Reorder tree according to order.

        Arguments:
            order {TupleTree, tuple, list} -- Enumerated structure used for reordering.

        Returns:
            TupleTree -- Same data reordered according to order.

        Example:
            a = tuplefy(list('abcd'), list('ef'))  # (('a', 'b', 'c', 'd'), ('e', 'f'))
            order = (0, (1, 2,), (5,))
            a.reorder(order)  # ('a', ('b', 'c'), ('f'))
        """
        order = tuplefy(order)
        flat_order = order.flatten()
        flat = self.flatten()
        return order.apply(lambda i: flat[i])

    def add_root(self):
        """Add a root node. Equivalent as 'TupleTree((self,))'."""
        return TupleTree((self,))

    def enumerate(self):
        """Creates list with numerated first element, and second element is previous data.

        Example:
            a = tuplefy(list('abcd'), list('ef'))  # (('a', 'b', 'c', 'd'), ('e', 'f'))
            a.enumerate()   # (([0, 'a'], [1, 'b'], [2, 'c'], [3, 'd']), ([4, 'e'], [5, 'f']))

        """
        counter = itertools.count()
        return self.apply(lambda x: [next(counter), x])


class _TupleTreeSlicer:
    def __init__(self, tuple_):
        self.tuple_ = tuple_

    def __getitem__(self, index):
        return self.tuple_.apply(lambda x: x[index])


# One can append to this to add inherited methods
_CONTAINERS = [TupleTree]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


with open('README.md', 'r') as readme_file:
    readme = readme_file.read()

requirements = [
    'numpy>=1.15.4',
    'pandas>=0.24.2',
    'matplotlib>=3.0.3',
    # 'torch>=1.1.0',
]


setup(
    name='torchtuples',
    version='0.2.1',
    description="Training neural networks in PyTorch",
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Haavard Kvamme",
    author_email='haavard.kvamme@gmail.com',
    url='https://github.com/havakv/torchtuples',
    packages=find_packages(include=['torchtuples']),
    include_package_data=True,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords='torchtuples',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.6'
)
